import java.net.Socket;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

/**
 * Wrapper-Klasse zum Verwalten einer Schicht-4-Verbindung (TCP).
 * 
 * @author Florian Janus
 * @version 2021-03-28
 */
public class Schicht4_TCP implements Schicht4
{
    private Socket clientSocket;
    private PrintWriter out;
    private BufferedReader in;
    private Thread reader;
    private boolean connected = false;

    

    public void verbindungAufbauen(String ip, int port) {
        if (connected) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Verbindung kann erst nach Verbindungsabau neu aufgabeut werden!");
            verbindungAbbauen();
        }
        try {
            clientSocket = new Socket(ip, port);
            out = new PrintWriter(clientSocket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
            connected = true;
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "Verbindung aufgebaut!");
            reader = new Thread(() -> { while (connected) {
                        try {
                            String line = in.readLine();
                            if (line != null) {
                                Terminal.printTerminal(Terminal.MSGTYPE.IN, line);
                            }
                        } catch (IOException ie) {
                            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Fehler beim Verarbeiten der eingehenden Daten!");
                        }
                    }});
            reader.start();
        } catch (IOException e) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Verbindungsaufbau fehlgeschlagen!");
        }
    }

    public void verbindungAbbauen() {
        try {
            in.close();
            out.close();
            clientSocket.close();
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "Verbindung abgebaut!");
        } catch (IOException e) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Verbindungsabbau fehlgeschlagen!");
        }
        finally {
            connected = false;
        }
    }

    public void nachrichtSenden(String msg) {
        if (!connected || out == null) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Kann Nachricht \"" + msg + "\" nicht senden, nicht verbunden!");
            return;
        }
        out.print(msg);
        out.flush();
        Terminal.printTerminal(Terminal.MSGTYPE.OUT, msg);
    }

    

}
