import java.net.Socket;
import java.net.InetAddress;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import java.net.UnknownHostException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.io.IOException;

/**
 * Wrapper-Klasse zum Verwalten einer verschluesselten Schicht-4-Verbindung (TCP + TLS).
 * 
 * @author Florian Janus 
 * @version 2021-03-28
 */
public class Schicht4_TCP_TLS implements Schicht4
{
    private Socket clientSocket;
    private SSLSession session;

    private PrintWriter out;
    private BufferedReader in;
    private Thread reader;
    private boolean connected = false;

    public void verbindungAufbauen(String ip, int port) {
        if (connected) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Verbindung kann erst nach Verbindungsabbau neu aufgebaut werden!");
            verbindungAbbauen();
        }
        try {

            SSLContext ctx = SSLContext.getInstance("TLS");
            X509TrustManager tm = new X509TrustManager() {

                    public void checkClientTrusted(X509Certificate[] xcs, String string) throws CertificateException {
                    }

                    public void checkServerTrusted(X509Certificate[] xcs, String string) throws CertificateException {
                    }

                    public X509Certificate[] getAcceptedIssuers() {
                        return null;
                    }
                };
            ctx.init(null, new TrustManager[]{tm}, null);
            SSLSocketFactory ssf = new NaiveSSLSocketFactory();
            clientSocket = ssf.createSocket(ip, port);

            session = ((SSLSocket) clientSocket).getSession();
            Certificate[] cchain = session.getPeerCertificates();
            for (int i = 0; i < cchain.length; i++) {
                Terminal.printTerminal(Terminal.MSGTYPE.NONE, "TLS: Zertifikat " + i + ": " + ((X509Certificate) cchain[i]).getSubjectDN());
            }
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "TLS: Peer host: " + session.getPeerHost());
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "TLS: Cipher: " + session.getCipherSuite());
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "TLS: Protocol " + session.getProtocol());

            out = new PrintWriter(clientSocket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
            connected = true;
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "Verbindung aufgebaut!");
            reader = new Thread(() -> { while (connected) {
                        try {
                            String line = in.readLine();
                            if (line != null) {
                                Terminal.printTerminal(Terminal.MSGTYPE.IN, line);
                            }
                        } catch (IOException ie) {
                            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Fehler beim Verarbeiten der eingehenden Daten!");
                        }
                    }});
            reader.start();

        } catch (Exception e) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Verbindungsaufbau fehlgeschlagen!");
            e.printStackTrace();
        }
    }

    public void verbindungAbbauen() {
        try {
            in.close();
            out.close();
            clientSocket.close();
            Terminal.printTerminal(Terminal.MSGTYPE.NONE, "Verbindung abgebaut!");
        } catch (IOException e) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Verbindungsabbau fehlgeschlagen!");
        }
        finally {
            connected = false;
        }
    }

    public void nachrichtSenden(String msg) {
        if (!connected || out == null) {
            Terminal.printTerminal(Terminal.MSGTYPE.ERROR, "Kann Nachricht \"" + msg + "\" nicht senden, nicht verbunden!");
            return;
        }
        out.print(msg);
        out.flush();
        Terminal.printTerminal(Terminal.MSGTYPE.OUT, msg);
    }
}

/*
 * Helferklasse: "Naive" Version einer SSLSocketFactory, die blind jedem TLS Endpunkt vertraut (=keine Pruefung der Zertifikatskette)
 */
class NaiveSSLSocketFactory extends SSLSocketFactory {
    SSLContext sslContext = SSLContext.getInstance("TLS");

    public NaiveSSLSocketFactory() throws NoSuchAlgorithmException, KeyManagementException, KeyStoreException, UnrecoverableKeyException {
        super();

        TrustManager tm = new X509TrustManager() {
                public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {
                }

                public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
                }

                public X509Certificate[] getAcceptedIssuers() {
                    return null;
                }
            };

        sslContext.init(null, new TrustManager[] { tm }, null);
    }

    @Override
    public String[] getSupportedCipherSuites() {
        return sslContext.getServerSocketFactory().getSupportedCipherSuites();
    }

    @Override
    public String[] getDefaultCipherSuites() {
        return sslContext.getServerSocketFactory().getDefaultCipherSuites();
    }

    @Override
    public Socket createSocket(InetAddress address, int port, InetAddress localAddress, int localPort) throws IOException {
        return sslContext.getSocketFactory().createSocket(address, port, localAddress, localPort);
    }

    @Override
    public Socket createSocket(InetAddress address, int port) throws IOException {
        return sslContext.getSocketFactory().createSocket(address, port);
    }

    @Override
    public Socket createSocket(String host, int port, InetAddress localHost, int localPort) throws IOException {
        return sslContext.getSocketFactory().createSocket(host, port, localHost, localPort);
    }

    @Override
    public Socket createSocket(String host, int port) throws IOException {
        return sslContext.getSocketFactory().createSocket(host, port);
    }

    @Override
    public Socket createSocket(Socket socket, String host, int port, boolean autoClose) throws IOException, UnknownHostException {
        return sslContext.getSocketFactory().createSocket(socket, host, port, autoClose);
    }

    @Override
    public Socket createSocket() throws IOException {
        return sslContext.getSocketFactory().createSocket();
    }
}
